import posixpath
import re
import time
import unittest
from urllib2 import urlopen, HTTPError
from xmlrpclib import Transport, ServerProxy, Fault

# Assumes the tracd is already running and pointed
# at a token server that allows decafbad.
TRAC_URL = 'http://code.timhatch.com:9000/launchpad4'
#TRAC_URL = 'http://localhost:8001/lpenv'

# If you're more patient than me, turn this off!
ANTSY = True


class LoginFailed(Exception): pass

def launchpad_login(server, token):
    url = posixpath.join(server, "launchpad-auth", token)
    try:
        r = urlopen(url)
    except HTTPError, e:
        raise LoginFailed(str(e))
    cookie = r.headers['Set-Cookie']

    m = re.match("(trac_auth=[a-zA-Z0-9]+)", cookie)
    if not m:
        raise LoginFailed("Auth cookie not provided")
    return m.group(1)


class CookieTransport(Transport):
    """xmlrpclib transport subclass to support providing a cookie with every
    request"""
    
    def __init__(self, cookie):
        self.cookie = cookie
        Transport.__init__(self)

    def send_host(self, connection, host):
        connection.putheader("Host", host)
        connection.putheader("Cookie", self.cookie)

trans = CookieTransport(launchpad_login(TRAC_URL, "decafbad"))
server = ServerProxy(posixpath.join(TRAC_URL, "xmlrpc"), transport=trans)

class BasicLaunchpadTests(unittest.TestCase):
    def test_butracker_version(self):
        (tver, pver, dupes) = server.launchpad.bugtracker_version()
        assert tver.startswith("0.11")
        assert pver == "0.1"
        assert dupes == False

    def test_timesnapshot(self):
        (tzname, localts, utcts) = server.launchpad.time_snapshot()
        self.assertEquals(type(tzname), str)
        utcoffset = localts - utcts
        #FIXME: hardcoded
        self.assertEquals(utcoffset % 60, 0)
        self.assertEquals(utcoffset, -5 * 3600)

        if not ANTSY:
            # Now we wait a bit, so the timestamps are a bit different
            time.sleep(2)
            (tzname2, localts2, utcts2) = server.launchpad.time_snapshot()
            # Make sure they change
            self.assertNotEqual(utcts, utcts2)
            self.assertNotEqual(localts, localts2)
        else:
            utcts2 = utcts
            localts2 = localts
        # Make sure they change at the same rate
        self.assertEquals(utcts2 - utcts, localts2 - localts)

    def test_numbugs(self):
        ts, numbugs = server.launchpad.num_bugs()
        self.assertEqual(numbugs, 3)

    def test_numbugs_component(self):
        ts, numbugs = server.launchpad.num_bugs("component2")
        self.assertEqual(numbugs, 1)

    def test_numbugs_bad_component(self):
        ts, numbugs = server.launchpad.num_bugs("badcomponent")
        self.assertEqual(numbugs, 0)

    def last_modified(self):
        ts, lm = server.launchpad.last_modified()

    def test_last_modified_badcomponent(self):
        #FIXME: verify the fault code
        self.assertRaises(Fault, server.launchpad.last_modified,
                          "badcomponent")

    def test_bug_components(self):
        ts, clist, defa = server.launchpad.bug_components()
        self.assertEquals(sorted(clist), ["component1", "component2"])
        self.assertEquals(defa, '')

    def test_most_recent_bugid(self):
        ts, bugid = server.launchpad.most_recent_bugid()
        self.assertEquals(bugid, 3)

    def test_bug_statuses(self):
        ts, statuses = server.launchpad.bug_statuses()
        #TODO: just the basic ones for now
        # full list on default0.11 is:
        # ['accepted', 'assigned', 'closed', 'new', 'reopened']
        self.assertEquals("new" in statuses, True)
        self.assertEquals("assigned" in statuses, True)

    def test_report_bug(self):
        pass

    def test_get_launchpad_bug(self):
        ts, bug = server.launchpad.get_launchpad_bug(1)
        self.assertEquals(bug, 123)

    def test_get_launchpad_bug_unset(self):
        ts, bug = server.launchpad.get_launchpad_bug(3)
        self.assertEquals(bug, 0)

    def test_get_launchpad_bug_badticket(self):
        #FIXME: verify the fault code
        self.assertRaises(Fault, server.launchpad.get_launchpad_bug, -1)

    def test_set_launchpad_bug(self):
        #FIXME: make this happen after the get_launchpad_bug tests.

        # This sets the bug on ticket #2 to something and then back
        server.launchpad.set_launchpad_bug(2, 0)
        ts, bug = server.launchpad.get_launchpad_bug(2)
        self.assertEquals(bug, 0)

        time.sleep(1) # Avoid key violation by waiting

        # Note that xmlrpclib won't let ones through that are too big.
        server.launchpad.set_launchpad_bug(2, 2**31-1)
        ts, bug = server.launchpad.get_launchpad_bug(2)
        self.assertEquals(bug, 2**31-1)

    def test_bug_info_oldlimit(self):
        ts, data = server.launchpad.bug_info(0, {'modified_since': 0})
        self.assertEquals(len(data), 3)

    def test_bug_info_newlimit(self):
        tomorrow = int(time.time()) + (3600 * 24)
        ts, data = server.launchpad.bug_info(0, {'modified_since': tomorrow})
        self.assertEquals(len(data), 0)

    def test_bug_info_component(self):
        ts, data = server.launchpad.bug_info(0, {'component': 'component1'})
        self.assertEquals([x['id'] for x in data], [2, 3])

    def test_bug_info_limits(self):
        ts, data = server.launchpad.bug_info(0, {'offset': 0, 'limit': 1})
        self.assertEquals(len(data), 1)
        self.assertEquals(data[0]['id'], 1)

    def test_bug_info_limits_with_offset(self):
        ts, data = server.launchpad.bug_info(0, {'offset': 1, 'limit': 1})
        self.assertEquals(len(data), 1)
        self.assertEquals(data[0]['id'], 2)
    
    def test_bug_info_bugs(self):
        ts, data = server.launchpad.bug_info(1, {'bugs': [2, 999]})
        self.assertEquals(len(data), 2)
        self.assertEquals(data[0]['id'], 2)
        self.assertEquals(data[1]['id'], 999)
        self.assertEquals(data[0]['status'], "new")
        self.assertEquals(data[1]['status'], "missing")

    def test_bug_info_bugs_missing_at_end(self):
        #TODO: seems that sqlite is ordering the IN() for us
        ts, data = server.launchpad.bug_info(0, {'bugs': [999, 4, 2, 5]})
        self.assertEquals([x['id'] for x in data], [2, 4, 5, 999])

    #TODO: test bug detail levels

    def test_bug_info_and_num_bugs(self):
        #FIXME: move after others
        ts, num_bugs = server.launchpad.num_bugs()
        ts, data = server.launchpad.bug_info(0)
        self.assertEquals(num_bugs, len(data))

        ts, data2 = server.launchpad.bug_info(0, {'offset': 0, 'limit':
            num_bugs})
        self.assertEquals(data, data2)

    def test_get_comments_empty(self):
        ts, comments = server.launchpad.get_comments([])
        self.assertEquals(comments, [])

    def test_get_comments_nonexist(self):
        ts, comments = server.launchpad.get_comments(['999-1'])
        self.assertEquals(len(comments), 1)
        self.assertEquals(comments[0]['type'], 'missing')
    
    def test_get_comments_real(self):
        ts, comments = server.launchpad.get_comments(['1-1'])
        self.assertEquals(comments[0]['comment'], 'A comment here')
        self.assertEquals(comments[0]['user'], 'launchpad')

    def test_get_comments_autonumber(self):
        #TODO
        pass


class ModifyingLaunchpadTests(unittest.TestCase):
    def test_report_bug(self):
        ts, bugid = server.launchpad.report_bug("Summary", "Desc", 123)

        ts, lpbug = server.launchpad.get_launchpad_bug(bugid)
        self.assertEquals(lpbug, 123)
    def test_add_comment(self):
        ts, bugid = server.launchpad.report_bug("Summary", "Desc", 0)
        ts, commentid = server.launchpad.add_comment(bugid, "Comment")


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(BasicLaunchpadTests, 'test'))
    suite.addTest(unittest.makeSuite(ModifyingLaunchpadTests, 'test'))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='suite')

