"""
Handles connecting Launchpad token-based authentication to Trac's cookie
auth method.

"""

# Copyright (C) 2008 Canonical, Ltd.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import posixpath
import re
import time
import urllib2

from trac.core import implements, Component, TracError
from trac.web.api import IRequestHandler
from trac.config import Option
from trac.util import hex_entropy
from trac.web.session import COOKIE_KEY

class LaunchpadAuthenticator(Component):
    """See module docstring"""

    implements(IRequestHandler)

    launchpad_server = Option('launchpad', 'auth_server',
        'https://launchpad.net/token/%s/+bugtracker-handshake',
        """Location of the official launchpad server, or one of your own for
        testing. See launchpad plugin docs for the interface.""")

    # IRequestHandler methods

    def match_request(self, req):
        return re.match('/launchpad-auth/([a-zA-Z0-9]+)$', req.path_info)

    def process_request(self, req):
        m = re.match('/launchpad-auth/([a-zA-Z0-9]+)$', req.path_info)
        input_token = m.group(1)

        server = self.launchpad_server
        if "%s" not in server:
            raise TracError("""Invalid [launchpad] auth_server
            configuration""")
        url = server % input_token

        self.log.debug("Verifying Launchpad token via %s" % url)
        try:
            # Providing data should force urllib2 to send a POST request
            token_req = urllib2.urlopen(url, 'a=1')
            # We don't care what textual data the server gives us, just
            # so long as the response code was 200 (i.e. doesn't raise
            # an error here).
            data = token_req.read()
        except Exception, e:
            #TODO: recognize specific faults outside of the plugin's
            # realm of being able to fix things:
            # - no route to host - 503
            # - connection refused - 503
            # - no ssl available - generic 500?
            self.log.debug("Invalid token: " + str(e))
            req.send_response(401) #Unauthorized
            req.end_headers()
            req.write("Invalid token")
            return

        self.log.debug("Token is valid.")

        # Note that since we don't store any session data, there's no
        # problem just generating a new auth cookie (sid) each time.
        sid = hex_entropy(24)

        db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.execute("INSERT INTO auth_cookie (cookie,name,ipnr,time)"
                       " VALUES (%s,%s,%s,%s)", (sid, "launchpad",
                                                 req.remote_addr,
                                                 int(time.time())))
        db.commit()

        self._purge_old_sessions()
        req.send_response(200)
        req.outcookie['trac_auth'] = sid
        req.outcookie['trac_auth']['path'] = '/'
        req.outcookie['trac_auth']['expires'] = 3600 * 24 #one day

        req.end_headers()
        req.write("OK")

    def _purge_old_sessions(self):
        cutoff = int(time.time()) - (3600 * 24) #one day

        db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.execute("DELETE FROM auth_cookie WHERE time < %s",
                       (cutoff, ))
        db.commit()
