"""
The meat of the Launchpad API for Trac is exposed through XML-RPC using the
existing XML-RPC plugin.
"""

# Copyright (C) 2008 Canonical, Ltd.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from datetime import datetime
import pkg_resources
import time
import xmlrpclib

have_genshi05 = False
try:
    from genshi.builder import tag
    from genshi.filters.transform import Transformer, ReplaceTransformation
    have_genshi05 = True
except ImportError:
    pass

import trac
from trac.config import Option
from trac.core import implements, Component, TracError
from trac.perm import IPermissionRequestor

from trac.ticket import model # so Component doesn't collide
from trac.ticket.api import TicketSystem
from trac.ticket.web_ui import TicketModule

from tracrpc.api import IXMLRPCHandler, expose_rpc

have_011 = False
try:
    from trac.util.datefmt import to_timestamp, to_datetime, utc
    from trac.web.api import ITemplateStreamFilter
    from trac.resource import ResourceNotFound
    have_011 = True
except ImportError:
    from trac.core import TracError as ResourceNotFound
    to_timestamp = None

def comment_ts(entry):
    if to_timestamp is None and 'raw_time' not in entry:
        raise xmlrpclib.Fault(1010, "Patch required to Trac core, see readme")
    elif to_timestamp is None:
        t = entry['raw_time']
    else:
        t = to_timestamp(entry['date'])
    return t

PERM = 'LAUNCHPAD_RPC'
LPUSER = "launchpad"

class LaunchpadRPC(Component):
    """
    Calls for Launchpad to sync with this Trac. See
    http://help.launchpad.net/Bugs/TracPlugin
    """

    implements(IXMLRPCHandler, IPermissionRequestor)
    if have_011:
        implements(ITemplateStreamFilter)

    dupe_field = Option('launchpad', 'dupe_field', 'duplicate_of',
            """If you use a custom ticket field for tracking duplicate
            tickets, put its name here.""")

    # IPermissionRequestor methods

    def get_permission_actions(self):
        yield 'LAUNCHPAD_RPC'

    # IXMLRPCHandler methods
    def xmlrpc_namespace(self):
        return 'launchpad'

    def xmlrpc_methods(self):
        yield (PERM, ((list,),), self.bugtracker_version)
        yield (PERM, ((list,),), self.time_snapshot)
        yield (PERM, ((list,), (int, str)), self.num_bugs)
        yield (PERM, ((list,), (int, str)), self.last_modified)
        yield (PERM, ((list,),), self.bug_components)
        yield (PERM, ((list,),), self.most_recent_bugid)
        yield (PERM, ((list,),), self.bug_statuses)
        yield (PERM, ((list, int),), self.get_launchpad_bug)
        yield (PERM, ((int, int, int),), self.set_launchpad_bug)
        yield (PERM, ((list, int),
                      (list, int, dict),
                     ),
                     self.bug_info)
        yield (PERM, ((list, list),), self.get_comments)
        yield (PERM, ((list, str, str, int),
                      (list, str, str, int, str),
                     ),
                     self.report_bug)
        yield (PERM, ((list, int, str),), self.add_comment)


    # Exported methods

    def bugtracker_version(self, req):
        """Returns version info (trac, lp plugin, dupes are known)"""
        tracver = trac.__version__
        lpver = pkg_resources.require("TracLaunchpad")[0].version
        
        knows_dupes = False
        dupe_field = self.config['launchpad'].get('dupe_field')
        for field in TicketSystem(self.env).get_ticket_fields():
            if field['name'] == dupe_field:
                knows_dupes = True
                break

        return (tracver, lpver, knows_dupes)

    def get_ts(self):
        """Internal helper for getting utc timestamp since it's used so
        much"""
        return int(time.time())

    def time_snapshot(self, req):
        """Returns (tz name, local ts, utc ts)"""
        utctime = int(time.time())
        if time.daylight:
            offset = time.altzone
        else:
            offset = time.timezone
        tz = time.tzname[0]
        self.log.debug("Timezone=%s, utc offset=%d" % (tz, -offset))

        localtime = utctime - offset
        return (tz, localtime, utctime)

    def num_bugs(self, req, component=None):
        ts = self.get_ts()
        base_query = "SELECT COUNT(id) FROM ticket"
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        if component is None:
            cursor.execute(base_query)
        else:
            cursor.execute(base_query + " WHERE component=%s", (component,))
        row = cursor.fetchone()
        if row[0] is None and component is not None:
            self.log.debug("No bugs in %s" % component)
            try:
                comp = model.Component(self.env, component, db=db)
            except TracError:
                self.log.debug("Component does not exist")
                raise xmlrpclib.Fault(1001, "Component no longer exists")
        return (ts, int(row[0] or 0))

    def last_modified(self, req, component=None):
        ts = self.get_ts()
        #TODO: change for attachments
        base_query = "SELECT MAX(changetime) FROM ticket"
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        if component is None:
            cursor.execute(base_query)
        else:
            cursor.execute(base_query + " WHERE component=%s", (component,))
        row = cursor.fetchone()
        if row[0] is None and component is not None:
            self.log.debug("No bugs in %s" % component)
            try:
                comp = model.Component(self.env, component, db=db)
            except TracError:
                self.log.debug("Component does not exist")
                raise xmlrpclib.Fault(1001, "Component no longer exists")
            row = [0]
        return (ts, int(row[0]))

    def bug_components(self, req):
        ts = self.get_ts()
        comps = [x.name for x in model.Component.select(self.env)]
        default_comp = self.env.config['ticket'].get('default_component', '')
        return (ts, comps, default_comp)

    def most_recent_bugid(self, req):
        ts = self.get_ts()
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.execute("SELECT MAX(id) FROM ticket")
        row = cursor.fetchone()
        if row[0] is None:
            # No bugs in the system?!
            return (ts, 0)
        return (ts, int(row[0]))

    def bug_statuses(self, req):
        ts = self.get_ts()
        statuses = [x.name for x in model.Status.select(self.env)]
        return (ts, statuses)

    def report_bug(self, req, summary, description, launchpad_bug,
                   component=None):
        ts = self.get_ts() #TODO: should this return the actual creation time?
        tic = model.Ticket(self.env)
        tic['summary'] = summary
        tic['description'] = description
        tic['status'] = 'new' #FIXME: check this against workflow etc
        tic['reporter'] = LPUSER
        if component:
            tic['component'] = component
        tic['launchpad_bug'] = str(launchpad_bug or '')
        tic.insert()

        try:
            tn = TicketNotifyEmail(self.env)
            tn.notify(tic, newticket=True)
        except Exception, e:
            self.log.exception("Notification failure for new Launchpad bug %d: %s"
                               % (tic.id, e))

        return (ts, tic.id)

    def get_launchpad_bug(self, req, bugid):
        ts = self.get_ts()
        try:
            tic = model.Ticket(self.env, bugid)
        except TracError:
            raise xmlrpclib.Fault(1001, "Ticket does not exist")
        # On 0.10, the dict contains an empty string
        v = tic.values.get('launchpad_bug', 0) or 0
        return (ts, int(v))

    def set_launchpad_bug(self, req, bugid, launchpad_bug):
        ts = self.get_ts()
        try:
            tic = model.Ticket(self.env, bugid)
        except TracError:
            raise xmlrpclib.Fault(1001, "Ticket does not exist")
        tic['launchpad_bug'] = str(int(launchpad_bug) or '')

        if not tic['launchpad_bug']:
            tic.save_changes(LPUSER, "Removing Launchpad bug entry.")
        else:
            link = "http://launchpad.net/bugs/%d" % bugid
            tic.save_changes(LPUSER, "See [%s Bug %d] in Launchpad." %
                                     (link, bugid))
        return ts

    def bug_info(self, req, level, criteria={}):
        """Returns matching bugs, with detail level specified"""
        ts = self.get_ts()
        buglist = criteria.get("bugs", None)
        modified_since = criteria.get("modified_since", None)
        offset = criteria.get("offset", None)
        limit = criteria.get("limit", None)
        component = criteria.get("component", None)

        if (offset is None and limit is not None) or \
           (offset is not None and limit is None):
            return xmlrpblib.Fault(1001, "offset and limit must be specified as a pair")
        if offset is not None:
            if not isinstance(offset, int) or \
               not isinstance(limit, int):
               return xmlrpclib.Fault(1005, "offset and limit must be ints")
        if modified_since is not None:
            if not isinstance(modified_since, int):
                return xmlrpclib.Fault(1002, "modified_since must be an int")
        if component is not None:
            if not isinstance(component, str):
                return xmlrpclib.Fault(1003, "component must be a str")
        if buglist is not None:
            try:
                buglist = map(str, map(int, buglist))
            except:
                return xmlrpclib.Fault(1004, "buglist must be integers")

        m = TicketModule(self.env)
        db = self.env.get_db_cnx()
        cursor = db.cursor()

        # Decide what bug ids should be returned.  Comments modify the
        # ticket.changetime
        query = "SELECT id FROM ticket %(where1)s" \
                "ORDER BY id %(offset)s"
        vars = {'where1': [], 'offset': ''}
        bindparams = []
        if buglist is not None:
            vars['where1'].append("id IN (%s)" % ','.join(map(str, map(int,
                buglist))))
        if modified_since is not None:
            vars['where1'].append("changetime >= %d" % modified_since)
        if component is not None:
            self.log.debug("Restrict to component %s" % component)
            vars['where1'].append("component=%s")
            bindparams.append(component)
        if offset is not None:
            vars['offset'] = "LIMIT %d OFFSET %d" % (limit, offset)

        for k in ('where1',):
            if not vars[k]: continue
            vars[k] = "WHERE %s " % " AND ".join(vars[k])

        self.log.debug("Newvars is %r" % vars)
        self.log.debug("Query is %s" % (query % vars))
        cursor.execute(query % vars, bindparams)

        def _cleanup(d):
            for k in d:
                if d[k] is None: d[k] = ''
            return d

        def _metadata_only(ticket_id, t):
            #TODO: just return most (all?) of t.values

            return {'id': ticket_id,
                    'status': t['status'] ,
                    'resolution': t['resolution'],
                    'priority': t['priority'],
                    'summary': t['summary'],
                    'description': t['description'], #TODO: clean up wiki
                    'reporter': t['reporter'], #TODO: look up email addr in
                                               #session info
                    'owner': t['owner'],
                    'component': t['component'],
                    'milestone': t['milestone'],
                   }

        def _comment_list(ticket_id, t):
            base = _metadata_only(ticket_id, t)
            comments = []
            for entry in m.grouped_changelog_entries(t, db):
                #TODO: figure out what to do with attachments.
                self.log.debug("Changelog: %r" % entry)
                if entry['permanent']:
                    comments.append("%d-%d" % (ticket_id, entry['cnum']))

            base['comments'] = comments
            return base

        def _comment_full(ticket_id, t):
            base = _metadata_only(ticket_id, t)
            comments = []
            for entry in m.grouped_changelog_entries(t, db):
                #TODO: figure out what to do with attachments.
                if entry['permanent']:
                    comments.append({'id': "%d-%d" % (ticket_id,
                                                      entry['cnum']),
                                     'type': "comment",
                                     'timestamp': comment_ts(entry),
                                     'comment': entry['comment']})
                #elif field == 'attachment':
                #    comments.append({'id': "%d-a%s" % (ticket_id, new),
                #                     'filename': new})

            base['comments'] = comments
            return base

        ret = []
        fns = {
            0: None, # Handled separately because it's the only one that
                     # doesn't fetch the full ticket data
            1: _metadata_only,
            2: _comment_list,
            3: _comment_full,
        }
        fn = fns[level]
        for ticket,  in cursor:
            ticket = int(ticket)
            if level == 0:
                ret.append({'id': ticket})
            else:
                t = model.Ticket(self.env, ticket) #FIXME: should probably pass db
                ret.append(_cleanup(fn(ticket, t)))

        if buglist:
            existing = set([x['id'] for x in ret])
            desired = set(map(int, buglist)) #FIXME: too many int/str convs.
            missing = desired - existing
            for i in missing:
                ret.append({'id': i, 'status': 'missing'})

        #TODO: PrivateTickets support
        return (ts, ret)

    def get_comments(self, req, comments):
        # Make sure the input is okay before hitting the db
        try:
            # Third element is the original... not even trimmed
            comment_list = [map(int, x.split('-', 1)) + [x] for x in comments]
        except TypeError:
            raise xmlrpclib.Fault(1001, "Invalid comment format")
        ts = self.get_ts()
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        ret = []
        
        fetch_list = {}
        for ticket, cnum, orig in comment_list:
            if ticket in fetch_list:
                fetch_list[ticket].append((cnum, orig))
            else:
                fetch_list[ticket] = [(cnum, orig)]

        def missing(orig):
            return {'type': 'missing', 'id': orig}

        def have(entry, orig):
            return {'type': 'comment', 'id': orig,
                    'user': entry['author'],
                    'comment': entry['comment'],
                    'timestamp': comment_ts(entry)}

        m = TicketModule(self.env)
        for ticket, comments in fetch_list.items():
            # If the ticket is invalid, all its comments are missing
            try:
                t = model.Ticket(self.env, ticket)
            except ResourceNotFound, e:
                for id, orig in comments:
                    ret.append(missing(orig))
                continue

            need = sorted(comments)
            for entry in m.grouped_changelog_entries(t, db):
                if entry['permanent']:
                    # These are entries that should be skipped, to catch up
                    while need and entry['cnum'] > need[0][0]:
                        id, orig = need.pop(0)
                        ret.append(missing(orig))
                    if not need: break
                    if entry['cnum'] == need[0][0]:
                        ret.append(have(entry, need[0][1]))
                        need.pop(0)
            # Requested entries > last real entry
            for id, orig in need:
                ret.append(missing(orig))
                
        return (ts, ret)
        
    def add_comment(self, req, bugid, comment):
        tic = model.Ticket(self.env, int(bugid))
        max_cnum = 0
        for entry in TicketModule(self.env).grouped_changelog_entries(tic, None):
            if entry['permanent']:
                max_cnum = entry['cnum']
        ret = tic.save_changes(LPUSER, comment, cnum=max_cnum+1)
        if not ret:
            raise Exception("Change was not saved")
        ts = self.get_ts()
        return (ts, "%s-%s" % (bugid, max_cnum+1))

    # ITemplateStreamFilter methods

    def filter_stream(self, req, method, filename, stream, data):
        if filename != 'ticket.html':
            # Not viewing a ticket, so skip the transform
            return stream

        lpbug = int(data['ticket'].values.get('launchpad_bug', '') or 0)
        if not lpbug:
            # Either the custom field isn't set, or it's empty
            return stream

        bug_link = tag.a("Bug %d" % lpbug,
                         href="http://launchpad.net/bugs/%d" % lpbug)
        transformer = Transformer('//td[@headers="h_launchpad_bug"]/text()').\
                      replace(bug_link)
        return stream | transformer
